TextArea = Core.class(Sprite)

TextArea.defaultFont = Font.getDefault()

function TextArea:init(font, text, sample,
align, width, letterspace, linespace)
	font = font or TextArea.defaultFont
	sample = sample or text
	local x, y, w, h = font:getBounds(sample)
	
	if not align then
		local child = TextField.new(font, text)
		self:addChild(child)
		if sample ~= text then
			local _
			x, _, w, _ = font:getBounds(text)
		end
		child:setPosition(-x, -y)
		
		self.font = font
		self.text = text
		self.sample = sample
		self.offsetY = y
		self.heightmul = h / Sprite.getHeight(self)
		return
	end
	
	local lines = {}
	
	if width then
		local words = {}
		local p1, p2 = sample:find"[ ]+"
		local indent = p2 and (" "):rep(p2 - p1 + 1)
		local text = text:gsub("\n", " \n ")
		for word in text:gmatch"([^ ]+)" do table.insert(words, word) end
		local line0 = indent
		for i, word in ipairs(words) do
			local newline = word == "\n"
			if newline then
				if words[i-1] == "\n" then
					table.insert(lines, "")
				elseif line0 then
					table.insert(lines, line0)
					line0 = indent
				end
			else
				local line = line0 and line0 .. " " .. word or word
				local _, _, lw, _ = font:getBounds(line)
				if lw > width then
					table.insert(lines, line0 or line)
					line0 = line0 and word
				else
					line0 = line
				end
			end
		end
		table.insert(lines, line0)
	else
		for line in text:gmatch"[^\r\n]+" do table.insert(lines, line) end
	end
	
	local t, xs, xs0, ws = {}, {}, {}, {}
	local w = 0
	local widest = 0
	local l = #lines
	
	for k,line in ipairs(lines) do
		local lx, _, lw, _ = font:getBounds(line)
		if lw > w then w = lw; widest = k end
		t[k], xs[k], ws[k], xs0[k] = TextField.new(font, line), lx, lw, 0
	end
	
	if letterspace then
		for i = 1, l do
			t[i]:setLetterSpacing(letterspace)
			ws[i] = t[i]:getWidth()
		end
		w = ws[widest]
	end
	
	if align == "R" then
		for i = 1, l do xs0[i] = w - ws[i] end
	elseif align == "C" then
		for i = 1, l do xs0[i] = 0.5 * (w - ws[i]) end
	elseif align == "J" then
		local nls = (letterspace or 0) + 1
		for i = 1, l - 1 do
			local nextline = lines[i+1]
			if nextline ~= "" and nextline:sub(1,1) ~= " " then
				local child = t[i]
				local w0 = ws[i]
				child:setLetterSpacing(nls)
				local w1 = child:getWidth()
				local s = math.floor((w - w0) / (w1 - w0))
				child:setLetterSpacing(s)
			end
		end
	end
	
	linespace = linespace or 0
	local h = h + linespace
	for i = 1, l do
		t[i]:setPosition(xs0[i] - xs[i], (i - 1) * h - y)
		self:addChild(t[i])
	end
	
	self.font = font
	self.text = text
	self.sample = sample
	self.align = align
	self.letterspace = letterspace or 0
	self.linespace = linespace or 0
	self.heightmul =  (l * h - linespace) / Sprite.getHeight(self)
	self.textcolor = 0x000000
	self.width = w
	self.lineheight = h - linespace
	self.offsetY = y
	self.widest = widest
	self.offsetXs = xs
end

function TextArea:setText(text)
	self.text = text
	if self.align then
		local _, child = next(self.__children)
		local textColor = child and child:getTextColor() or 0x000000
		table.foreach(self.__children, function(_, child)
			child:removeFromParent()
		end)
		TextArea.init(self, self.font, self.text, self.sample,
			self.align, self.width, self.letterspace, self.linespace)
		if textColor ~= 0x000000 then self:setTextColor(textColor) end
	else
		local _, child = next(self.__children)
		local h0 = Sprite.getHeight(self)
		child:setText(text)
		self.heightmul = self.heightmul * h0 / Sprite.getHeight(self)
	end
end

function TextArea:getText(color)
	return self.text
end

function TextArea:setSample(sample)
	self.sample = sample or self.text
	local x, y, w, h = self.font:getBounds(self.sample)
	if self.align then
		self.offsetY, self.lineheight = y, h
		self:setLineSpacing(self.linespace)
	else
		local _, child = next(self.__children)
		child:setY(-y)
		self.offsetY = y
		self.heightmul = h / Sprite.getHeight(self)
	end
end

function TextArea:getSample()
	return self.sample
end

function TextArea:setLetterSpacing(letterspace)
	self.letterspace = letterspace
	if not self.align then
		local _, child = next(self.__children)
		child:setLetterSpacing(letterspace)
	elseif self.__children then
		local child = self:getChildAt(self.widest)
		child:setLetterSpacing(letterspace)
		self.width = child:getWidth()
		local w = self.width
		local align = self.align
		
		local t, xs, ws, xs0 = {}, self.offsetXs, {}, {}
		table.foreach(self.__children, function(_, child)
			child:setLetterSpacing(letterspace)
			local i = self:getChildIndex(child)
			t[i], ws[i], xs0[i] = child, child:getWidth(), 0
		end)
		local l = #t
		
		if align == "R" then
			for i = 1, l do xs0[i] = w - ws[i] end
		elseif align == "C" then
			for i = 1, l do xs0[i] = 0.5 * (w - ws[i]) end
		elseif align == "J" then
			local nls = (letterspace or 0) + 1
			for i = 1, l - 1 do
				local nextline = t[i+1]:getText()
				if nextline ~= "" and nextline:sub(1,1) ~= " " then
					local child = t[i]
					local w0 = ws[i]
					child:setLetterSpacing(nls)
					local w1 = child:getWidth()
					local s = math.floor((w - w0) / (w1 - w0))
					child:setLetterSpacing(s)
				end
			end
		end
		
		for i = 1, l do t[i]:setX(xs0[i] - xs[i]) end
	end	
end

function TextArea:getLetterSpacing(color)
	return self.letterspace
end

function TextArea:setLineSpacing(linespace)
	self.linespace = linespace
	if not self.__children then return end
	local h = self.lineheight + linespace
	local y = self.offsetY
	local l = self:getNumChildren()
	for i = 1, l do
		local child = self:getChildAt(i)
		child:setY((i - 1) * h - y)
	end
	self.heightmul = (l * h - linespace) / Sprite.getHeight(self)
end

function TextArea:getLineSpacing()
	return self.linespace
end

function TextArea:setAlignment(align)
	self.align = align
	self:setLetterSpacing(self.letterspace)
end

function TextArea:getAlignment()
	return self.align
end

function TextArea:setTextColor(color)
	self.textcolor = color
	if self.__children then
		table.foreach(self.__children, function(_, child)
			child:setTextColor(color)
		end)
	end
end

function TextArea:getTextColor()
	return self.textcolor
end

function TextArea:getHeight()
	return self.heightmul * Sprite.getHeight(self)
end